/* This is the definition file for the framebuffer class */

//Frame buffer is implemented as a doubly linked list with a tail and a 
//front header

#include "framebuffer.h"
#include "constants.h"
#include <iostream>
using namespace std;

//constructor
framebuffer::framebuffer(link *next, int size)
{
   capacity = size;
   numframe = 0;
   nextlink = next;
   nextfinished = -1;
   tailheader = new frames;
   tailheader = NULL;
   frontheader = new frames;
   frontheader = NULL;
}

/************************************************************
  inserts a new frame at the end of the frame buffer
  Input: pointer to the frame to be inserted
         pointer to the eventlist
         pointer to the buffer in which the frame is to be inserted
         current time
  Output: None
*******************************************************************/
void framebuffer::include(frame *newframe, eventlist* header,framebuffer 
*sourcebuff, double current_time)
{
     if(numframe != capacity) //buffer has space
     { 
       frames *next_frame = new frames;
       next_frame->framebuffer_node = newframe;
       next_frame->nextframe=tailheader;
       if(tailheader != NULL)
         tailheader->prevframe = next_frame;
       else
       {
          next_frame->prevframe = NULL;
          frontheader = next_frame;
       }
       tailheader = next_frame;
       next_frame->prevframe = NULL;
       numframe++;
       
     }
   
      sourcebuff->extract(sourcebuff, header, current_time);
}

/*****************************************************************
  Extracts the frame at the front of the buffer to the link
  Input: pointer to the buffer from which the frame has to be extracted
         pointer to the eventlist
         current time
  Output: None
*******************************************************************/
void framebuffer::extract(framebuffer *sourcebuff, eventlist * 
neweventlist, double current_time)
{
   double TD, prev_linktime;
   if(frontheader != NULL)
   {
      
      int framesize = frontheader->framebuffer_node->tlength;
      double SD = (double)framesize/(nextlink->linespeed);
     
       if(frontheader->framebuffer_node->type == 0)
        prev_linktime = prevdata_linktime;
      else
        prev_linktime = acklinktime;

      if(prev_linktime == 0)
        TD = SD + current_time;
      else if(current_time < prev_linktime)
      {
        TD = SD + prev_linktime;
      }
      else
        TD = SD + current_time;

      event *BuffertoLinkevent = new event();
      BuffertoLinkevent->scheduleevent(TD, frontheader->framebuffer_node, 
	2, NULL, NULL, NULL, sourcebuff, NULL, nextlink);
      
      PQE *newPQE = new PQE(TD, 2, BuffertoLinkevent);
      neweventlist->enqueue(*newPQE);
      frames *sent_frame = frontheader;
      sent_frame->framebuffer_node = NULL;
      frontheader = frontheader->prevframe;
      
      if(frontheader == NULL)
      {
         tailheader = NULL;
         
      }
      
      sent_frame->prevframe = NULL;
      delete sent_frame;
      sent_frame = NULL;
       numframe--;

      if(prev_linktime == prevdata_linktime)
        prevdata_linktime = TD;
      else
        acklinktime = TD;

   }
}

//Makes the frame buffer empty. Called when stopping the simualtor

void framebuffer::destroyframes()
{
  frames *header;
  header = tailheader;
  frames *ptr;
  while(header!=NULL)
  {
    ptr = header->nextframe;
    delete header;
    header = NULL;
    header = ptr;
  }
  nextlink = NULL;
  delete tailheader;
  delete frontheader;
  tailheader = NULL;
  frontheader = NULL;
}


