/* The file contains the definition files for the node class */
#include <ctime>
#include <stdlib.h>
#include "node.h"
#include "struct_h.h"
#include <iostream>
#include "constants.h"
using namespace std;

//constructor
node::node(int window, double timeout)
{
   W = window;
   source_hacked = 0;
   source_hsent = 1;
   dest_hacked = 1;
   sent_frame_count = 1;
   nextbuffer = NULL;
   nacked_frame_seq = 0;
   state = Normal;
   nacked_frame_to = 0;
   sent_unacked_framecount = 1;
   frame_TO = timeout;
}

//Intializes the nextbuffer pointer of the framebuffer class
void node::setnextbuff(framebuffer *buffer)
{
   nextbuffer = buffer;
}

/*****************************************************************
This function is invoked when a data frame arrives at the destination node

Input:  frame -> pointer to the arrived frame
                 pointers to source and destination nodes
                 pointer to the eventlist
                 current_time
********************************************************************/

void node::receiveframeatdest(frame *receivedframe, node 
*destn, eventlist * neweventlist, double current_time)
{
  if(receivedframe->status == OK)
  {
   if(receivedframe->type == Data)
   {
       if(receivedframe->dataseq == (dest_hacked+1))
       {  //received next expected frame, creating ack
	  frame * newframe = generateframe(ACK, receivedframe->dataseq);
          event *newevent = new event();

          double t = current_time;
	  newevent->scheduleevent(t, newframe, 1,destn, NULL, nextbuffer, NULL, NULL, NULL);

         //create new PQE node for the event: Frame handed from node to 
         //buffer
          PQE *newPQE = new PQE(t, 1, newevent);
          neweventlist->enqueue(*newPQE);      //insert the newly
                                              //created in the eventlist
          dest_hacked = receivedframe->dataseq;
 	  
       }
       if(receivedframe->dataseq > (dest_hacked+1)) 
       {  //received out-of-order data frame, create and send NACK

          frame * newframe = generateframe(NACK, (dest_hacked+1));
          event *newevent = new event();

	  double t = current_time;
          newevent->scheduleevent(t, newframe, 1,destn, NULL, nextbuffer, 
            NULL, NULL, NULL);

          //create new PQE node for event: FRame handed from node to 
          //buffer
          PQE *newPQE = new PQE(t, 1, newevent);
          neweventlist->enqueue(*newPQE);
	  
       }
    }
   }
   else //arrived frame is damaged, send NACK saying expected 
        //(dest_hacked+1)
   {
	  frame * newframe = generateframe(NACK, (dest_hacked+1));
          event *newevent = new event();

          double t = current_time;
          newevent->scheduleevent(t, newframe, 1,destn, NULL, nextbuffer, 
	     NULL, NULL, NULL);
          PQE *newPQE = new PQE(t, 1, newevent);
          neweventlist->enqueue(*newPQE);
         
    }
}

/**********************************************
     ACK or NACK arrived at source node
Input: frame -> pointer to the arrived frame
       pointers to source and destination nodes
       pointer to the eventlist
       current_time
************************************************/

void node::receiveframeatsource(frame *receivedframe, node *source, 
eventlist * neweventlist, double current_time)
{
  double newoffset;
  if(receivedframe->status == OK)
  {
   if(receivedframe->type == ACK)
   {
     if(receivedframe->ackseq > source_hacked)
     {
        state = Normal;
        sent_unacked_framecount -= (receivedframe->ackseq - source_hacked);
	source_hacked = receivedframe->ackseq;
        
        //received ack, send data frames (source_hsent+1) to
        //(source_hsent+W)
        int count = (source->W) - sent_unacked_framecount;
        for(int i=(source_hsent+1); i<=(source_hsent+count); i++)
	{
	   frame *new_frame = generateframe(Data, i);
	  
	   sent_frame_count++;
	   sent_unacked_framecount++;

           //create a node to buffer event for the newly generated frame
           event *newevent1 = new event();

           //calculation of offset time for successively generated frames

           double SD_data = (double)(new_frame->tlength)/LINESPEED;
           double SD_ack = (double)(receivedframe->tlength)/LINESPEED;
           if(SD_data < SD_ack)
             newoffset = SD_data/W;
           else
             newoffset = SD_ack/W;
           if(offset > newoffset)
             offset = newoffset;

           double t;
           if(i == (source_hsent+1))
             t = current_time;
           else
             t = current_time + offset;
           
           newevent1->scheduleevent(t, new_frame, 1,source, NULL, nextbuffer, NULL, NULL, NULL);
           PQE *newPQE = new PQE(t, 1, newevent1);
           neweventlist->enqueue(*newPQE);
          
	   
	   //creating a TO event for the newly generated frame
	   event * newevent2 = new event();
	  
           double to_time = t + frame_TO;
         
	   newevent2->scheduleevent(to_time, new_frame, 0, source, source, 
	     NULL, NULL, NULL, NULL);
	   PQE *newPQE1 = new PQE(to_time, 0, newevent2);
	   neweventlist->enqueue(*newPQE1);
	   
	   current_time = t;
        }
        source_hsent = source_hsent + count;
        
    }
   }
   else
   {
    if(state == Normal)
    {
     if(receivedframe->ackseq == (source_hacked+1))
     {      //received NACK from destination
        nacked_frame_seq = receivedframe->ackseq;
        state = Recovery;
      //send again frames from (source_hacked+1) to (source_hacked+W)
	   for(int i=(source_hacked+1); i<=(source_hacked+source->W); i++)
           {
           frame *new_frame = generateframe(Data, i);
           sent_frame_count++;

           //create node to frame event
           event *newevent1 = new event();

           //calculation of offset time for successively generated frames

           double SD_data = (double)(new_frame->tlength)/LINESPEED;
           double SD_ack = (double)(receivedframe->tlength)/LINESPEED;
           if(SD_data < SD_ack)
             newoffset = SD_data/W;
           else
             newoffset = SD_ack/W;
           if(offset > newoffset)
             offset = newoffset;

           double t;
           if(i == (source_hacked+1))
             t = current_time;
           else
             t = current_time + offset;
           newevent1->scheduleevent(t, new_frame, 1,source, NULL, nextbuffer, NULL, NULL, NULL);
           PQE *newPQE = new PQE(t, 1, newevent1);
           neweventlist->enqueue(*newPQE);
           
           //creating a TO event
	   event * newevent2 = new event();
           double SD = (double)new_frame->tlength/LINESPEED;
           double to_time = t + frame_TO;
           
           if(i == receivedframe->ackseq)
             nacked_frame_to = to_time;
           
           newevent2->scheduleevent(to_time, new_frame, 0, source, source,
             NULL, NULL, NULL, NULL);
           PQE *newPQE1 = new PQE(to_time, 0, newevent2);
           neweventlist->enqueue(*newPQE1);
           current_time = t;
        }
        //changing source_hsent to source_hacked+W
        source_hsent = source_hacked + W;
        sent_unacked_framecount = W;
     }
   }
  }
 }
}
 
/*******************************************************************
   This function is called when a time out event occurs 
   Input: sequence number of the frame for which the time out occurs
          pointer to the eventlist
          pointer to the source
          pointer to the PQE node which contains the generated TO event
          current time
   output: None
************************************************************************/
void node::timeout(int seqno, eventlist *head, node *source, PQE *PQEptr, 
double current_time)
{
    double newoffset;
    if(source_hacked<seqno)
    { //true TO
       PQEptr->ptr->timeout = true;

      //time out occurred, create frames from (source_hacked+1) to 
      //(source_hacked+W)
       for(int i=(source_hacked+1); i<=(source_hacked+W); i++)
        {
           frame *new_frame = generateframe(Data, i);
           sent_frame_count++;
        
           //creating a node to buffer event
           event *newevent1 = new event();

           //calculation of offset time for successively generated frames

           double SD_data = (double)(new_frame->tlength)/LINESPEED;
           double SD_ack = (double)(DATAFRAME_SIZE)/LINESPEED;
           if(SD_data < SD_ack)
             newoffset = SD_data/W;
           else
             newoffset = SD_ack/W;
           if(offset > newoffset)
             offset = newoffset;

           double t;
           if(i == (source_hacked+1))
             t = current_time;
           else
             t = current_time + offset;
           newevent1->scheduleevent(t, new_frame, 1,source, NULL, nextbuffer, NULL, NULL, NULL);
           PQE *newPQE = new PQE(t, 1, newevent1);
           head->enqueue(*newPQE);
          
           //creating a TO event
	   event * newevent2 = new event();
        
           double to_time = t + frame_TO;
          
           newevent2->scheduleevent(to_time, new_frame, 0, source, source,
             NULL, NULL, NULL, NULL);
           PQE *newPQE1 = new PQE(to_time, 0, newevent2);
           head->enqueue(*newPQE1);
           current_time = t;
        }
        //changing source_hsent
        source_hsent = source_hacked+W;
    }
}

double node::myrand()
{
   return ((double)rand()*1.0/(double)RAND_MAX);
}

/************************************************************
        create new frames
	Input: Type of the frame to be created
               Its sequence number
        output: Pointer to the created frame
*************************************************************/
frame * node::generateframe(Frame_Type type, int seqno)
{
    frame *newframe = new frame;
    int dlength;

    switch(type)
    {
	case Data:
           newframe->settype(Data);
           //dlength = (int)(myrand()*2.0*MSG_SIZE);
           dlength = DATA_LENGTH;
           break;
	case ACK:
	   newframe->settype(ACK);
           //dlength = 10;
           dlength = ACKDATA_LENGTH;
	   break;
	case NACK:
	   newframe->settype(NACK);
           //dlength = 10;
	   dlength = ACKDATA_LENGTH;
	   break;
    };
    
    newframe->setstatus(OK);
    newframe->setdlength(dlength);
    newframe->sethlength(HEADER_LENGTH);
    int tlength = HEADER_LENGTH + dlength;
    newframe->settlength(tlength);
    if(type == Data)
    {
      newframe->setdataseq(seqno);
      newframe->setackseq(0);
    }
    else
    {
      newframe->setackseq(seqno);
      newframe->setdataseq(0);
    }
    return newframe;
}

