#ifndef _NODE_H_
#define _NODE_H_
/*This file the class definition for the node class */
#include "link.h"
#include "framebuffer.h"
#include "PQE.h"
#include "struct_h.h"
#include "constants.h"
#include "frame.h"

class eventlist;    /*see PQE.h*/
class framebuffer;  /*see framebuffer.h*/
class frame;        /*see frame.h*/

class node  
{
public:
int W ;            /* Window size: That is the number of outstanding 
		      frames */
int source_hacked ; /* At source: highest framenumber acknowledgement 
		       received.
		       At destination: it is unused(as the data traffic 
			is assumed to be unidirectional) */
int source_hsent  ; /* At source: highest framenumber sent, thus far. 
                       At destination: it is unused (for unidirectional 
			data traffic*/
Node_State state;

double frame_TO;       /*Time out interval of a data frame */
int dest_hacked   ; /* At destination: highest framenumber acknowledgement 
		       sent.
		       At source: it is not used */
int sent_frame_count;/*At source: total number of frames sent by source 
		       counting retransmissions
                       At destination: it is not used */

int sent_unacked_framecount; /*At source: it is the number of frames sent
                               but not yet acknowleged
                               At destination: it is not used */
int nacked_frame_seq;       /*At source: it is the frame sequence for 
			      which a NACK was received from destn */

double nacked_frame_to;
framebuffer * nextbuffer ; /* pointer to the framebuffer this node hands 
				frames to. */

/* and some functions */
public:

node(int w, double timeout);    //constructor
/* w is the window size of the node i.e. the number of outstanding frames 
*/

void setnextbuff(framebuffer *);
/* sets the nextbuffer pointer of the node object. It is the buffer to
   which the node hands in frames
*/

void receiveframeatsource(frame *arrivedframe, node *arrivednode,eventlist 
*header, double current_time);
/* When a frame arrives at the source node, receiveframeatsource processes 
the frame to see if it is a ACK/NACK frame and acts accordingly
(generates new frames or frames for retransmission along with events for 
them)
 "arrivedframe" is the pointer to the arrived frame
 "arrivednode" is the pointer to the node at which the frame arrives.
 "header" is the pointer to the eventlist(There is only one eventlist).
 "current_time" is the time at which the frame arrives in the node.
*/

void receiveframeatdest(frame *arrivedframe, node *arrival_node,  
eventlist *header,double current_time);
/* When a frame arrives at the destination node, receiveframeatdest 
processes the frame to see if it is damaged/ok and checks the sequence 
number of the frame and generates an ACK if it is received the frame 
it was expecting and it is ok or a NACK if it received a frame that 
was out-of-sequence or a damaged frame
 "arrivedframe" is the pointer to the arrived frame
 "arrival_node" is the pointer to the node at which the frame arrives.
 "header" is the pointer to the eventlist(There is only one eventlist).
 "current_time" is the time at which the frame arrives in the node.
*/

void timeout(int seqno, eventlist *header, node *eventnode, PQE *eventPQE, 
double current_time);
/* When timeout occurs for a frame, the timeout method checks to see if it 
was a true(frame has really gotten timedout) or a false(ACK for the frame 
has already been received) timeout and generates frames and events for a 
true timeout and does nothing if it is a false timeout.
 "seqno" sequence number of the timedout frame
 "header" is the pointer to the eventlist(There is only one eventlist).
 "eventnode" is the node in which timeout occurred
 "current_time" is the time at which the timeout occurred
*/

double myrand(); //generates and returns a random number

//generates and returns a new frame
frame * generateframe(Frame_Type type, int seqno);
/* generateframe generates and returns a new frame
   "type" is the type of frame generated ->Data/ACK/NACK
   "seqno" incase of data frame, it is the sequence number of the frame
           incase of ACK, it is the frame sequence number it acknowledges
           incase of NACK, it is the sequence number of the frame the 
		destination was expecting to receive
*/

};

#endif

