/*This is the main file for the simulator which checks for events and *
  invokes the corresponding event handlers.                           */

#include <iostream>
#include "link.h"
using namespace std;
#include <stdlib.h>

main()
{
  //Initialization of source and destination nodes
  node *snode = new node(W, TIMEOUT);
  node *dnode = new node(0, 0);

  //Initialization of forward and reverse link
  link *flink = new link(dnode);
  link *rlink = new link(snode);

  //Initialization of source and destination frame buffers
  framebuffer *sbuffer = new framebuffer(flink, 10000);
  framebuffer *dbuffer = new framebuffer(rlink, 10000);

  //Initialization of eventlist
  eventlist *eventptr = new eventlist();
  
  int s_hsent, TO;
  TO = 0;
 
  cout<<"Simulator started ........\n";
 cout<<"Line speed of link: "<<LINESPEED<<"bits/sec,\tPropagation delay: "; 
  cout<<PROP_DELAY<<"secs\n";
  cout<<"Loss probability: "<<PL<<",\tDamage probability: "<<PD<<"\n";
  cout<<"OK probability: "<<POK<<"\n";
  cout<<"Source Window size: "<<W<<",\tFrame size: ";
  cout<<(DATAFRAME_SIZE/8)<<"bytes\n";
  cout<<"Timeout interval: "<<TIMEOUT<<"\n";
  srand(101);

  //generating the first ACK frame for sequence 1 and creating a 
  //first link-to-node event and adding it in the eventlist to start the 
  //chain of events

  frame *newframe = dnode->generateframe(ACK, 1);
  double t = 0;
  
  event *newevent = new event();
  newevent->time = t;
  newevent->sourcetype = Link;
  newevent->desttype = Node;
  newevent->sourcenode = NULL;
  newevent->destnode = snode;
  newevent->sourcebuffer = NULL;
  newevent->destbuffer = NULL;
  newevent->sourcelink = rlink;
  newevent->destlink = NULL;
  newevent->thisframe = newframe;
  newevent->dataseq = 0;
  PQE *newPQE = new PQE(t, 3, newevent);
  PQEnode *newnode = new PQEnode;
  newnode->PQE_node = newPQE;
  newnode->next = NULL;
  eventptr->first = newnode; 
  
  //Entering an infinite while loop to check for events
  while(true)
  {
    PQE * PQE_event;
    
    //Simulator will halt when the sent_frame_count becomes > than 250
    if(snode->sent_frame_count <= FINAL_FRAME)
    {
    if(eventptr->first == NULL) //eventlist is empty
      continue;
    else
    {
      PQE_event = eventptr->dequeue();
      switch(PQE_event->type) 
      {
        case 0:
          //Time out event occurred
         if((snode->nacked_frame_seq == PQE_event->ptr->dataseq)&& 
		(snode->nacked_frame_to == PQE_event->time))
         {
           snode->state = Normal;
           snode->nacked_frame_seq = 0;
           snode->nacked_frame_to = 0;
           TO = 0;
         }
         if(snode->state == Normal)
         {
          if(TO == 0)
	  {
	   PQE_event->ptr->TO_handler(eventptr, snode, PQE_event, 
		PQE_event->time);
           if(PQE_event->ptr->timeout)
           {
            s_hsent = snode->source_hsent;
	    TO = 1;

            //Will get the output only if there is true time out
	    cout<<"\nTime out event occurred for frame sequence ";
	    cout<<PQE_event->ptr->dataseq<<" at time ";
	    cout<<PQE_event->time<<"secs\n";
           }
	  }
          if(s_hsent == PQE_event->ptr->dataseq)
	  {
            s_hsent = 0;
	    TO = 0;
          }
         }
	  delete PQE_event->ptr;
          PQE_event = NULL;
          delete PQE_event;
          PQE_event = NULL;
          break;
	case 1:
          //Frame given from node to buffer
          if(PQE_event->ptr->thisframe->type == 0)
          {
	    cout<<"\nData frame sequence "<<PQE_event->ptr->dataseq;
            cout<<" arrived at buffer 1 at time = "; 
            cout<<PQE_event->time<<"secs\n";
            
            PQE_event->ptr->Node2Buff_handler(eventptr, sbuffer, 
		PQE_event->time);
          }
          else
	  {
	    if(PQE_event->ptr->thisframe->type == 1)
	     cout<<"\nACK frame acknowledging sequence "; 
            else
             cout<<"\nNACK frame expecting sequence ";
            cout<<PQE_event->ptr->thisframe->ackseq;
            cout<<" arrived at buffer 2 at time = ";
            cout<<PQE_event->time<<"secs\n";
            PQE_event->ptr->Node2Buff_handler(eventptr, dbuffer, 
		PQE_event->time);
	  }
          delete PQE_event->ptr;
          PQE_event = NULL;
          delete PQE_event;
          PQE_event = NULL;
          break;
        case 2:
          //Frame given from buffer to link
          if(PQE_event->ptr->thisframe->type == 0)
	  {
	    cout<<"\nData frame sequence "<<PQE_event->ptr->dataseq;
            cout<<" arrived at link 1 at time = ";
            cout<<PQE_event->time<<"secs\n";
            PQE_event->ptr->Buffer2Link_handler(eventptr, flink, 
		PQE_event->time);
	  }
          else
          {
	   if(PQE_event->ptr->thisframe->type == 1)
	    cout<<"\nACK frame for sequence ";
	   else
            cout<<"\nNACK frame expecting sequence ";
           cout<<PQE_event->ptr->thisframe->ackseq;
	   cout<<" arrived at link 2 at time = ";
           cout<<PQE_event->time<<"secs\n";
           PQE_event->ptr->Buffer2Link_handler(eventptr, rlink, 
		PQE_event->time);
	  }
          delete PQE_event->ptr;
          PQE_event = NULL;
          delete PQE_event;
          PQE_event = NULL;
          break;
        case 3:
         // Frame given from link to node
          if(PQE_event->ptr->thisframe->type == 0)
	  {
            cout<<"\nData frame sequence "<<PQE_event->ptr->dataseq;
            cout<<" arrived at node 2 at time = ";
            cout<<PQE_event->time;
            cout<<"secs, Status ";
            if(PQE_event->ptr->thisframe->status == 0)
               cout<<"OK.\n";
            else
               cout<<"Damaged.\n";

	  }
	  else 
          {
	    if(PQE_event->ptr->thisframe->type == 1)
		cout<<"\nACK frame for sequence ";
	    else
		cout<<"\nNACK frame expecting sequence ";
            cout<<PQE_event->ptr->thisframe->ackseq;
	    cout<<" arrived at node 1 at time = ";
	    cout<<PQE_event->time;
            cout<<"secs, Status ";
            if(PQE_event->ptr->thisframe->status == 0)
               cout<<"OK.\n";
            else
               cout<<"Damaged.\n";

	  }
	  PQE_event->ptr->Link2Node_handler(snode, dnode, eventptr, 
		PQE_event->time);
          delete PQE_event->ptr;
          PQE_event = NULL;
          delete PQE_event;
          PQE_event = NULL;
          break;
        
      };
    }
   }
   else
   {
       
       cout<<"Stopping the simulator....\n";
       sbuffer->destroyframes();
       //cout<<"Source buffer destroyed\n";
       dbuffer->destroyframes();
       //cout<<"Destn buffer destroyed\n";
       eventptr->destroyeventlist();
       //cout<<"Eventlist destroyed\n";
       delete snode;
       snode = NULL;
       delete dnode;
       dnode = NULL;
       delete flink;
       flink = NULL;
       delete rlink;
       rlink = NULL;
       delete sbuffer;
       sbuffer = NULL;
       delete dbuffer;
       dbuffer = NULL;
       delete eventptr;
       eventptr = NULL;
       exit(0);
   }

  }  
}
